#' @title Format covey data
#'
#' This function creates data frames with the format and columns needed for integrated likelihood modeling with random effect count models
#' The data frames are \code{dis.object} which is at the resolution of the detections and \code{glmm.data} which is at the resolution of the counts at the sampler (or counts in the distance bins).

#' @param data data frame from user; 
#' @param w truncation distance in units specified in \code{dis.unit}
#' @param sampler "lines" (default) or "points" 
#' @param dis.unit units of distances (perpendicular for lines and radial for points)
#' @param sampler.units required field for line transects (default = \code{dis.unit})
#' @param exact TRUE for exact distance measurements, FALSE for interval distance data
#' @param cutpoints required if exact = FALSE: list of cutpoints in same units as \code{dis.unit}, needs to be of length I+1 for I cutpoints with cutpoints[1] = 0 and cutpoints[I+1] = w
#' @param dis.cov optional list of covariates that may be included in MCDS model (expects a list of column names from \code{data} where covariates are found)
#' @param count.cov required list of covariates that may be included in count model (expects the column names from data where covariates are found, no NA's allowed, only one value for each visit to a sampler allowed)
#' @param dis.factor.cov list of covariates to be considered as factors
#' @param count.factor.cov list of covariates to be considered as factors
#' @param conditional TRUE creates 1 column of for detections and 1 column for individuals in glmm.data. If False, creates 1 column for detections and individuals for each distance bin (only applicable for interval distance data)

#' @details 
#' 
#' For \code{data} the following columns are expected:
#' 
#'  \code{$gr.id} identifiers for groups (grouping factor for random effect); 
#'  
#'  \code{$smp.id} identifiers for points/lines; 
#'  
#'  \code{$vis.id} identifiers for individual visits to the same sampler (in case of repeat visits); different visits need to have different lables for each unique smp.id;
#'  
#'  \code{$det.id} identifiers for detections; 
#'  
#'  \code{$line.length} for line transects only; 
#'  
#'  \code{$distance} exact or midpoints for intervals for interval data; 
#'  
#'  \code{$size} cluster size; 
#'  
#'  covariates for detection or count model (no NA's allowed);
#'  
#' Also for \code{data} the following rows are expected: 
#' 
#' 1 row for each detection; 
#' 
#' if no detection at a visit to a sampler: 1 row with NA's for \code{$det.id}, \code{$distance} and \code{$size}

#' @examples
#' covey.data<-create.data(covey,500,"points","m","m",F,dis.cov=c("Type","State"),count.cov=c("Type","State","JDctr"), dis.factor.cov = c("Type","State"), count.factor.cov = c("Type","State"), conditional=T)

create.data <-
  function(data, w, sampler = "lines", dis.unit, sampler.unit = dis.unit, 
           binned, cutpoints = NULL, dis.cov = NULL, count.cov, dis.factor.cov = NULL,
           count.factor.cov = NULL, conditional = T) {
    
    data$smp.vis<-paste(data$gr.id,data$smp.id,data$vis.id,sep="_") # creates a unique identifier for each visit to each sampler of each group
    
    data.object<-NULL                                               # object that will be returned by this function
    data.object$w <- w                                              # the truncation distance
    data.object$sampler <- sampler                                  # lines or points
    data.object$dis.unit <- dis.unit                                # units for distances (cm, m, km)
    data.object$sampler.unit <- sampler.unit                        # units for $line.length (cm, m, km) 
    data.object$binned <- binned
    data.object$cutpoints <- cutpoints
    
    data.object$dis.cov <- dis.cov
    dis.cov.fac <- array(NA,length(dis.cov))
    for (u in 1:length(dis.cov)){
      dis.cov.fac[u] <- ifelse(is.na(match(dis.cov[u],dis.factor.cov))==F,TRUE,FALSE)
    }
    data.object$dis.cov.fac<-dis.cov.fac
    
    data.object$count.cov<-count.cov
    count.cov.fac<-array(NA,length(count.cov))
    for (u in 1:length(count.cov)){
      count.cov.fac[u]<-ifelse(is.na(match(count.cov[u],count.factor.cov))==F,TRUE,FALSE)
    }
    data.object$count.cov.fac<-count.cov.fac
    
    data.object$conditional<-conditional
    
    # $dis.object = a data.frame containing information required for detection function model L_{y|z}
    if(sampler == "lines")  {dis.object<-data[which(is.na(data$det.id)==F),c("smp.vis","det.id","smp.id","gr.id","vis.id","line.length","distance","size",dis.cov)]}
    if(sampler == "points") {dis.object<-data[which(is.na(data$det.id)==F),c("smp.vis","det.id","smp.id","gr.id","vis.id","distance","size",dis.cov)]  }
    # truncating at w
    if(binned==T){
      bins<-which.bin(data$distance,cutpoints)
      data$bin<-bins[[1]]
      data[,c("distbegin","distend")]<-bins[[2]]
    }
    dis.object<-data[which(data$distance<=w),]
    
    # adding to the data.object
    data.object$dis.object<-data.frame(dis.object)
    
    # $glmm.data a data frame containing the information for the random effect count model L_n
    if(conditional==T){
      data.object$glmm.data<-create.glmm.data(data, w, sampler, count.cov, count.cov.fac) }
    else{
      data.object$glmm.data<-create.glmm.data.unconditional(data, w, sampler, count.cov, count.cov.fac)           
    }
    data.object
  }


#' @title Uniform prior probabilities
#' 
#' @details
#' This function returns the summed log of the uniform prior probabilities 
#' 
#' @param coef vector of coefficient values
#' @param lo vector of lower boundaries for uniform distributions
#' @param up vector of upper boundaries for uniform distributions 
#' 
#' @details 
#' In contrast to using *sum(log(dunif(...)))*, the \code{l.prior} function returns 
#' a very small value for the log of the prior probability if the value is 
#' outside the boundaries of the distribution. This prevents the acceptance of 
#' a newly proposed parameter value if it is outside the boundaries. 
#' In contrast, using *log(dunif(...))* returns *-Inf* if the parameter value is outside 
#' the boundaries which may cause the algorithm to crash. 
#' 
l.prior<-function(coef,lo,up){
  l.u.int<-log(dunif(coef,lo,up))
  if(any(abs(l.u.int)==Inf))l.u.int<- -100000
  sum(l.u.int)
}


#' @title Evaluating the probability density of observed distances using the hazard-rate model for points
#'
#' The probability density of observed distances for a given distance x1, given scale and 
#' shape parameter sigma1 and shape1 and given truncation distance w is obtained using: 
#' f.haz.function.pt(x1,sigma1,shape1)/integrate(f.haz.function.pt,0,w,sigma1,shape1)$value
#' 
#' @return 
#' Probability density
#' 
#' @param dis exact radial distance measurement
#' @param sigma scale parameter 
#' @param shape shape parameter

f.haz.function.pt<-function(dis,sigma,shape) {
  f <- dis*(1-exp(-(dis/sigma)^(-shape)))
  f
}


#' @title Evaluating the detection probability using the hazard-rate model
#'
#' Evaluates the detection probability for a given distance dis, given scale and 
#' shape parameters sigma1 and shape1 
#' 
#' @return 
#' Detection probability
#' 
#' @param dis exact radial distance measurement
#' @param sigma scale parameter 
#' @param shape shape parameter

g.haz.function.pt<-function(dis,sigma,shape) {
  g <- 1-exp(-(dis/sigma)^(-shape))
  g
}



#' @title Matching strings of numbers against rows in a matrix
#' 
#' This function matches a string of numbers against a matrix
#' and returns the row number of the matrix that matches the string exactly
#' 
#' @param xmod a string of numbers
#' @param model.matrix A matrix of numbers

match.function<-function(xmod,model.matrix){
  nmodel<-nrow(model.matrix)
  npar<-ncol(model.matrix)
  is.match.or.not<-array(NA,nmodel)
  for (i in 1:nmodel){
    is.match.or.not[i]<-sum(xmod==model.matrix[i,])}
  result<-which(is.match.or.not==npar)
  return(result)
}

#' @title Likelihood for the detection model component fitted to the covey data
#' 
#' @details
#' This function is set up to fit a harzard-rate detection model to the covey data. 
#'  
#' @return 
#' Log-likelihood
#' 
#' @param p Current model parameter values of this specific model (see details)
#' @param datax Data object formatted with \code{create.data}: covey.data

covey.hr.detmodel.log.lik<-function(p, datax){
  ## Part 1: setting up the parameter values p  
  # the detection model
  scale<-exp(p[1])     # the scale intercept 
  shape<-exp(p[2])     # the shape parameter on the log scale
  
  ## Part 2: the likelihood component pertaining to the detection model
  # calculating the f(y) for each observed distances
  le<-nrow(datax$dis.object)
  fe<-numeric(le)
  # calculating the f(y) for each observation 
  norm.const<-integrate(f.haz.function.pt,0,datax$w,scale,shape)$value
  for (e in 1:le){
    fe[e]<-f.haz.function.pt(datax$dis.object$distance[e],scale,shape)/norm.const
  }
  # the sum of the log(f(y))
  log.e<-sum(log(fe))
  return(log.e)
}


#' @title Combined likelihood for a simple model fitted to the covey data
#' 
#' @details
#' This function is set up to fit a model to the covey data with a hazard-rate detection model
#' containing no covariates and the count model containing one fixed effect covariate, Type.
#' 
#' @return 
#' Log-likelihood
#' 
#' @param p Current model parameter values of this specific model (see details)
#' @param raneff Current values for the random effect coefficients
#' @param datax Data object formatted with \code{create.data}: covey.data
#' 
#' @examples 
#' covey.int.log.lik(p = c(sig.0,sha.0,int.0,typ.0,std.ran.0), raneff, covey.data)

covey.int.log.lik<-function(p, raneff, datax){
  ## Part 1: setting up the parameter values p for the detection function and count models 
  # the detection model
  scale<-exp(p[1]) # the scale intercept 
  shape<-exp(p[2])     # the shape parameter on the log scale
  # the count model 
  int<-p[3]            # the intercept
  typ<-c(0,p[4])       # coefficient for Type level "TREAT" (level "CONTROL" is absorbed in the intercept)
  sd.ran<-exp(p[5]) # the random effect standard deviation on the log scale
  
  # Part 2: the likelihood component pertaining to the detection model
  # calculating the f(y) for each observed distances
  le<-nrow(datax$dis.object)
  fe<-numeric(le)
  # calculating the f(y) for each observation (note that the truncation distance is stored in datax$w)
  norm.const<-integrate(f.haz.function.pt,0,datax$w,scale,shape)$value
  for (e in 1:le){
    fe[e]<-f.haz.function.pt(datax$dis.object$distance[e],scale,shape)/norm.const
  }
  # the sum of the log(f(y))
  log.e<-sum(log(fe))
  
  # Part 3 of covey.int.log.lik(): the likelihood component pertaining to the count model  
  # setting up indices for the factor covariate Type and random effect coefficients
  Type0<-match(datax$glmm.data$Type,sort(unique(datax$glmm.data$Type)))
  gr.id<-sort(unique(datax$glmm.data$gr.id))
  Ran0<-match(datax$glmm.data$gr.id,gr.id)
  # log of effective area (here: same for each visit to a point due to global det model)
  l.efa<-log(norm.const*pi*2)
  # calculate the log-Poisson density for each count 
  lambda<-exp(int + typ[Type0] + raneff[Ran0] + l.efa)
  dpois.y<-dpois(datax$glmm.data$detections,lambda)
  logdpois.y<-sum(log(dpois.y))
  # calculate the log of the normal density for each random effect coefficient
  log.dnorm.raneff<-sum(log(dnorm(raneff,0,sd.ran)))
  # adding up the likelihood components
  log.lik<-logdpois.y + log.dnorm.raneff + log.e
  # the function return
  return(log.lik)
}

#' @title Updating random effects coefficients using a combined likelihood for a simple model fitted to the covey data
#' 
#' @details
#' This function is set up to fit a model to the covey data with a hazard-rate detection model
#' containing no covariates and the count model containing one fixed effect covariate, Type (factor with 2 levels).
#' 
#' @return 
#' Updated random effects coefficients
#' 
#' @param p Current model parameter values of this specific model (see details)
#' @param raneff Current values for the random effect coefficients
#' @param datax Data object formatted with \code{create.data}: covey.data
#' 
#' @examples 
#' covey.int.log.lik.raneff.update(p = c(sig.0,sha.0,int.0,typ.0,std.ran.0), raneff, covey.data)

covey.int.log.lik.raneff.update<-function(p,raneff,datax){
  # setting up the covariates
  scale<-exp(p[1])
  shape<-exp(p[2])
  int<-p[3]
  typ<-c(0,p[4])
  sd.ran<-exp(p[5])
  
  # log of effective area (here: same for each visit to a point due to global det model)
  norm.const<-integrate(f.haz.function.pt,0,datax$w,scale,shape)$value
  l.efa<-log(norm.const*pi*2)
  # setting up indices for the factor covariate Type and random effect coefficients
  Type0<-match(datax$glmm.data$Type,sort(unique(datax$glmm.data$Type)))
  gr.id<-sort(unique(datax$glmm.data$gr.id))
  Ran0<-match(datax$glmm.data$gr.id,gr.id)
  # calculate the log-Poisson density for each count 
  lambda<-exp(int + (typ[Type0]) + raneff[Ran0] + l.efa)
  #lambda<-as.numeric(unlist(lambda))
  # update each coefficient 
  for (r in 1:length(gr.id)){
    u<-rnorm(1,0,0.2)
    yr<-which(datax$glmm.data$gr.id==gr.id[r])
    dpois.y.new<-dpois(datax$glmm.data$detections[yr],lambda[yr]*exp(u))
    num<-sum(log(dpois.y.new))+log(dnorm(raneff[r]+u,0,sd.ran))
    dpois.y.cur<-dpois(datax$glmm.data$detections[yr],lambda[yr])
    den<-sum(log(dpois.y.cur))+log(dnorm(raneff[r],0,sd.ran))
    A<-min(1,exp(num-den))
    V<-runif(1)
    if(V<=A){raneff[r]<-raneff[r]+u}
  }
  # return updated coefficients
  raneff
}

#' @title Combined likelihood for a simple model fitted to the covey data
#' 
#' @details
#' This function is set up to fit a model to the covey data with a hazard-rate detection model
#' containing no covariates and the count model containing two fixed effect covariates, Type (factor with two levels) and JDctr.
#' 
#' @return 
#' Log-likelihood
#' 
#' @param p Current model parameter values of this specific model (see details)
#' @param raneff Current values for the random effect coefficients
#' @param datax Data object formatted with \code{create.data}: covey.data
#' 
#' @examples 
#' covey.rj.int.log.lik(p = c(sig.0,sha.0,int.0,typ.0,std.ran.0), raneff, covey.data)

covey.rj.int.log.lik<-function(p, raneff, datax){
  ## Part 1: setting up the parameter values p for the detection function and count models 
  # the detection model
  scale<-exp(p[1])     # the scale intercept 
  shape<-exp(p[2])     # the shape parameter on the log scale
  # the count model 
  int<-p[3]            # the intercept
  typ<-c(0,p[4])       # coefficient for Type level "TREAT" (level "CONTROL" is absorbed in the intercept)
  day<-p[5]
  sd.ran<-exp(p[6])    # the random effect standard deviation on the log scale
  
  # Part 2: the likelihood component pertaining to the detection model
  # calculating the f(y) for each observed distances
  le<-nrow(datax$dis.object)
  fe<-numeric(le)
  # calculating the f(y) for each observation (note that the truncation distance is stored in datax$w)
  norm.const<-integrate(f.haz.function.pt,0,datax$w,scale,shape)$value
  for (e in 1:le){
    fe[e]<-f.haz.function.pt(datax$dis.object$distance[e],scale,shape)/norm.const
  }
  # the sum of the log(f(y))
  log.e<-sum(log(fe))
  
  # Part 3 of covey.int.log.lik(): the likelihood component pertaining to the count model  
  # setting up indices for the factor covariate Type and random effect coefficients
  Type0<-match(datax$glmm.data$Type,sort(unique(datax$glmm.data$Type)))
  gr.id<-sort(unique(datax$glmm.data$gr.id))
  Ran0<-match(datax$glmm.data$gr.id,gr.id)
  # log of effective area (here: same for each visit to a point due to global det model)
  l.efa<-log(norm.const*pi*2)
  # calculate the log-Poisson density for each count 
  lambda<-exp(int + typ[Type0] + covey.data$glmm.data$JDctr*day + raneff[Ran0] + l.efa)
  dpois.y<-dpois(datax$glmm.data$detections,lambda)
  logdpois.y<-sum(log(dpois.y))
  # calculate the log of the normal density for each random effect coefficient
  log.dnorm.raneff<-sum(log(dnorm(raneff,0,sd.ran)))
  # adding up the likelihood components
  log.lik<-logdpois.y + log.dnorm.raneff + log.e
  # the function return
  return(log.lik)
}


#' @title Updating random effects coefficients using a combined likelihood for a simple model fitted to the covey data
#' 
#' @details
#' This function is set up to fit a model to the covey data with a hazard-rate detection model
#' containing no covariates and the count model containing two fixed effect covariates, Type (factor with 2 levels) and JDctr.
#' 
#' @return 
#' Updated random effects coefficients
#' 
#' @param p Current model parameter values of this specific model (see details)
#' @param raneff Current values for the random effect coefficients
#' @param datax Data object formatted with \code{create.data}: covey.data
#' 
#' @examples 
#' covey.rj.int.log.lik.raneff.update(p = c(sig.0,sha.0,int.0,typ.0,day.0,std.ran.0), raneff, covey.data)

covey.rj.int.log.lik.raneff.update<-function(p,raneff,datax){
  # setting up the covariates
  scale<-exp(p[1])
  shape<-exp(p[2])
  int<-p[3]
  typ<-c(0,p[4])
  day<-p[5]
  sd.ran<-exp(p[6])
  
  # log of effective area (here: same for each visit to a point due to global det model)
  norm.const<-integrate(f.haz.function.pt,0,datax$w,scale,shape)$value
  l.efa<-log(norm.const*pi*2)
  # setting up indices for the factor covariate Type and random effect coefficients
  Type0<-match(datax$glmm.data$Type,sort(unique(datax$glmm.data$Type)))
  gr.id<-sort(unique(datax$glmm.data$gr.id))
  Ran0<-match(datax$glmm.data$gr.id,gr.id)
  # calculate the log-Poisson density for each count 
  lambda<-exp(int + (typ[Type0]) + covey.data$glmm.data$JDctr*day + raneff[Ran0] + l.efa)
  #lambda<-as.numeric(unlist(lambda))
  # update each coefficient 
  for (r in 1:length(gr.id)){
    u<-rnorm(1,0,0.2)
    yr<-which(datax$glmm.data$gr.id==gr.id[r])
    dpois.y.new<-dpois(datax$glmm.data$detections[yr],lambda[yr]*exp(u))
    num<-sum(log(dpois.y.new))+log(dnorm(raneff[r]+u,0,sd.ran))
    dpois.y.cur<-dpois(datax$glmm.data$detections[yr],lambda[yr])
    den<-sum(log(dpois.y.cur))+log(dnorm(raneff[r],0,sd.ran))
    A<-min(1,exp(num-den))
    V<-runif(1)
    if(V<=A){raneff[r]<-raneff[r]+u}
  }
  # return updated coefficients
  raneff
}

#' @title Format distance data into count data
#'
#' This function converts distance sampling data containing records of detections, into count data, where each visit to a sampler are the records.
#' It is an internal function that is called by the \code{create.data} function.
#'
#' @param data data frame from user;
#' expects the following columns:
#'  \code{$det.id} (identifiers for detections);
#'  \code{$smp.id} (identifiers for points/lines);
#'  \code{$gr.id} (identifiers for groups (grouping factor for random effect));
#'  \code{$vis.id} (identifiers for individual visits to the same sampler (in case of repeat visits));
#'  \code{$line.length} (for line transects only);
#'  \code{$distance} (exact or midpoints for intervals for interval data);
#'  \code{$size} (cluster size);
#'  covariates for count model (no NA's allowed);
#' expects the following rows:
#'  1 row for each detection;
#'  if no detection at a visit to a sampler: 1 row with NA's for \code{$det.id}, \code{$distance} and \code{$size}
#'  
#' @param w truncation distance in units specified in \code{dis.unit}
#' @param sampler "lines" (default) or "points"
#' @param count.cov required list of covariates that may be included in count model (expects the column names from \code{data}, no NA's allowed, only one value for each visit to a sampler allowed)
#' @param count.cov.fac list of covariates in \code{count.cov} that are factors

create.glmm.data<-function(data, w, sampler, count.cov, count.cov.fac) {
  data$smp.vis<-paste(data$gr.id,data$smp.id,data$vis.id,sep="_") # creates a unique identifier for each visit to each sampler of each group
  glmm.raw<-data[,c("smp.vis","det.id","smp.id","gr.id","vis.id","distance","size",count.cov)]
  if(sampler=="lines"){glmm.raw$line.length<-data[,"line.length"]}
  # truncating the data at w while making sure that effort lines remain in place
  truncated<-which(glmm.raw$distance>w)
  if(length(truncated)>0){
    glmm.raw[truncated,c("det.id","distance","size")]<-NA
  }
  # list of unique smp.vis (each unique smp.vis will have one row in glmm.data)
  smp.vis<-unique(data$smp.vis)
  
  if(sampler=="lines"){glmm.data<-matrix(NA,length(smp.vis),7+length(count.cov))
  glmm.data<-data.frame(glmm.data)
  colnames(glmm.data)<-c("smp.vis","smp.id","gr.id","vis.id","detections","individuals","line.length",count.cov)
  glmm.data$smp.vis<-smp.vis
  for (s in 1:length(smp.vis)){
    rows<-which(glmm.raw$smp.vis==smp.vis[s])
    glmm.data$smp.id[s]<-paste(glmm.raw$smp.id[rows[1]])
    glmm.data$gr.id[s]<-paste(glmm.raw$gr.id[rows[1]])
    glmm.data$vis.id[s]<-paste(glmm.raw$vis.id[rows[1]])
    glmm.data$detections[s]<-length(which(is.na(glmm.raw$det.id[rows])==F))
    glmm.data$individuals[s]<-sum(glmm.raw$size[rows],na.rm=T)
    glmm.data$line.length[s]<-paste(glmm.raw$line.length[rows[1]])
    for (t in 1:length(count.cov)){
      if(count.cov.fac[t]==T){
        glmm.data[s,count.cov[t]]<-paste(glmm.raw[rows[1],count.cov[t]])
      }
      else{
        glmm.data[s,count.cov[t]]<-glmm.raw[rows[1],count.cov[t]]
      }
    }
  }
  }
  
  if(sampler=="points"){glmm.data<-matrix(NA,length(smp.vis),6+length(count.cov))
  glmm.data<-data.frame(glmm.data)
  colnames(glmm.data)<-c("smp.vis","smp.id","gr.id","vis.id","detections","individuals",count.cov)
  glmm.data$smp.vis<-smp.vis
  for (s in 1:length(smp.vis)){
    rows<-which(glmm.raw$smp.vis==smp.vis[s])
    glmm.data$smp.id[s]<-paste(glmm.raw$smp.id[rows[1]])
    glmm.data$gr.id[s]<-paste(glmm.raw$gr.id[rows[1]])
    glmm.data$vis.id[s]<-paste(glmm.raw$vis.id[rows[1]])
    glmm.data$detections[s]<-length(which(is.na(glmm.raw$det.id[rows])==F))
    glmm.data$individuals[s]<-sum(glmm.raw$size[rows],na.rm=T)
    for (t in 1:length(count.cov)){
      if(count.cov.fac[t]==T){
        glmm.data[s,count.cov[t]]<-paste(glmm.raw[rows[1],count.cov[t]])
      }
      else{
        glmm.data[s,count.cov[t]]<-glmm.raw[rows[1],count.cov[t]]
      }
    }
  }
  }
  return(glmm.data)
}



